/*  $Id: SCBalloonScrollArrow.java,v 1.6 2004/02/05 09:29:34 phonohawk Exp $

    "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
    Copyright (C) 2001-2004 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
import com.apple.cocoa.foundation.*;
import com.apple.cocoa.application.*;
import util.*;

public class SCBalloonScrollArrow extends NSView {
    public static final float SCROLL_STEP = 10f;
    public static final int DEFAULT_WIDTH = 15;
    public static final int DEFAULT_HEIGHT = 15;
    
    SCSafetyBalloonController controller;
    NSClipView target;
    int type; // SCBalloonSkinServer.ARROW_UP or SCBalloonSkinServer.ARROW_DOWN
    NSImage image;
    NSColor bg_color,arrow_color; // image == null̎ɗLB
    NSTimer scroll_timer;
    boolean isVisible = false;
    boolean isDrawn = false;

    NSImage offscreen;
    private NSRect[] rectlist_bounds = null;
    
    public SCBalloonScrollArrow(SCSafetyBalloonController controller,int type) {
	// ftHgB摜gȂB
	super(new NSRect(0,0,DEFAULT_WIDTH,DEFAULT_HEIGHT));
	
	this.controller = controller;
	this.target = controller.getTextScrollView().contentView();
	this.type = type;
	this.image = null;
	this.bg_color = NSColor.colorWithCalibratedRGB(0.5f,0.5f,0.5f,0.8f);
	this.arrow_color = NSColor.colorWithCalibratedRGB(1.0f,1.0f,1.0f,0.8f);
	this.offscreen = new NSImage(new NSSize(DEFAULT_WIDTH, DEFAULT_HEIGHT));
    }

    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("type: ").append(
	    type == SCBalloonSkinServer.ARROW_UP ? "up" : "down");
	if (image != null) {
	    buf.append("; has image");
	}
	buf.append("; ").append(isVisible ? "visible" : "invisible");
	buf.append("; location: ").append(frame().x()).append(", ").append(frame().y());
	buf.append("; size: ").append(frame().width()).append(", ").append(frame().height());
	return buf.toString();
    }
    
    public void cleanUp() {
	controller = null;
    }
    
    public void setImage(NSImage image) {
	// imagenullłǂB摜̎gp𒆎~B
	this.image = image;
	
	if (image == null) {
	    NSSize size = new NSSize(DEFAULT_WIDTH, DEFAULT_HEIGHT);
	    setFrameSize(size);
	    offscreen.setSize(size);
	}
	else {
	    setFrameSize(image.size());
	    offscreen.setSize(image.size());
	}
	rectlist_bounds = new NSRect[] {bounds()};
    }
	
    protected void updateOffscreen() {
	// OŕKoffscreenbN/鎖B
	if (SCFoundation.DEBUG) {
	    // g\
	    //NSColor.redColor().set();
	    //NSGraphics.frameRect(bounds());
	}

	/*
	  `rŝ̂͂₤ȏꍇ́AisVisibletruełĂ
	  Â`UȂ΂ȂȂB
	  At@`lĂꍇłB
	*/
	
	if (!isVisible) {
	    if (isDrawn) {
		NSColor.clearColor().set();
		NSGraphics.fillRectList(rectlist_bounds);
		
		isDrawn = false;
	    }
	    return;
	}
	
	if (image == null) {
	    bg_color.set();
	    NSGraphics.fillRectList(rectlist_bounds);
	    
	    arrow_color.set();
	    if (type == SCBalloonSkinServer.ARROW_UP) {
		NSPoint arrowCapPoint = new NSPoint(width() / 2f,height()-1);
		NSBezierPath.strokeLineFromPoint(
		    new NSPoint(width() / 2f,1),
		    arrowCapPoint); // |
		NSBezierPath.strokeLineFromPoint(
		    new NSPoint(width() / 3f,height() / 2f),
		    arrowCapPoint); // /
		NSBezierPath.strokeLineFromPoint(
		    new NSPoint(width() * (2f / 3f),height() / 2f),
		    arrowCapPoint); // \
	    }
	    else {
		NSPoint arrowCapPoint = new NSPoint(width() / 2f,1);
		NSBezierPath.strokeLineFromPoint(
		    new NSPoint(width() / 2f,height()-1),
		    arrowCapPoint); // |
		NSBezierPath.strokeLineFromPoint(
		    new NSPoint(width() / 3f,height() / 2f),
		    arrowCapPoint); // \
		NSBezierPath.strokeLineFromPoint(
		    new NSPoint(width() * (2f / 3f),height() / 2f),
		    arrowCapPoint); // /
	    }
	}
	else {
	    image.compositeToPoint(
		NSPoint.ZeroPoint, NSImage.CompositeSourceOver);
	}
	isDrawn = true;
    }
    public void drawRect(NSRect r) {
	offscreen.lockFocus();
	updateOffscreen();
	offscreen.unlockFocus();

	if (isVisible) {
	    offscreen.compositeToPointFromRect(
		r.origin(), r, NSImage.CompositeSourceOver);
	}
    }

    public void mouseDown(NSEvent event) {
	scroll_timer = new NSTimer(
	    0.03,
	    this,
	    new NSSelector("scroll",new Class[0]),
	    new Object[0],
	    true);
	SCFoundation.sharedFoundation().getRunLoopOfMainThread().
	    addTimerForMode(scroll_timer,NSRunLoop.DefaultRunLoopMode);
    }
    
    public void mouseUp(NSEvent event) {
	scroll_timer.invalidate();
    }
    
    protected void scroll() {
	controller.scroll(
	    type == SCBalloonSkinServer.ARROW_UP ?
	    SCROLL_STEP : -1 * SCROLL_STEP);
    }
    
    public void updateVisibility() {
	NSRect current_visible_rect = target.documentVisibleRect();
	float current_vert_point = current_visible_rect.y();
	
	if (type == SCBalloonSkinServer.ARROW_UP) {
	    float max_vert_point = target.documentRect().height() -
		current_visible_rect.height();
	    setVisibility(current_vert_point < max_vert_point);
	}
	else {
	    setVisibility(current_vert_point > 0f);
	}
    }
    
    protected void setVisibility(boolean isVisible) {
	if (this.isVisible != isVisible) {
	    this.isVisible = isVisible;
	    //setNeedsDisplay(true);

	    // {҂͂̂₤ɃJIX̗͂Ȃǎg͂ȂĂsetNeedsDisplay(true)œȂ̂
	    // ǂ]桂Ał̓A[ɟkĂ܂ӁB̃R[ĥAppKit̂sB
	    // ҂AppKitł̌ۂNȂȂÃIJg߂鎖B
	    // MacOS X 10.3.2 (2004/02/05)
	    NSTimer display_timer = new NSTimer(
		0.0,
		this,
		new NSSelector("display",new Class[] {}),
		new Object[] {},
		false);
	    SCFoundation.sharedFoundation().getRunLoopOfMainThread().addTimerForMode(
		display_timer, NSRunLoop.DefaultRunLoopMode);
	}

	if (isVisible) {
	    controller.setIgnoresMouseEvents(false);
	}
    }
    
    public float width() {
	return frame().width();
    }
    
    public float height() {
	return frame().height();
    }
    
    public void setFrameOrigin(float x,float y) {
	setFrameOrigin(new NSPoint(x,y));
    }
}
