/*
    "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
    Copyright (C) 2001 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
import java.io.*;
import java.util.*;

public class SCESDictionary
{
    SCEseShiori eseshiori;
    
    Hashtable root; // key : (String)Gg , value : (SCESNormalEntry)Ggf[^
    Hashtable special; // key : (String)^Cv , value : (Vector)SCESSpecialEntry̏W
    // content : (SCESSpecialEntry)Gg
    
    public SCESDictionary(SCEseShiori eseshiori)
    {
	this.eseshiori = eseshiori;
	root = new Hashtable();
	special = new Hashtable();
    }
    
    public void loadDict(File dicfile)
    {
	// dicfile : [ht@C
	try
	{
	    BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(dicfile),"SJIS"));
	    String currentEntryName = null;
	    StringBuffer currentLine = new StringBuffer();
	    while (true)
	    {
		String line = br.readLine();
		if (line == null)
		{
		    // EOF
		    if (currentLine.length() > 0 && currentEntryName != null) // flush
		    {
			addDataToEntry(currentLine.toString(),currentEntryName);
			currentLine.setLength(0);
		    }
		    break;
		}
		if (line.length() == 0) continue;
		if (line.startsWith("//")) continue;
		
		if (line.charAt(0) == '#')
		{
		    if (currentLine.length() > 0 && currentEntryName != null) // flush
		    {
			addDataToEntry(currentLine.toString(),currentEntryName);
			currentLine.setLength(0);
		    }
		    currentEntryName = line.substring(1);
		}
		else if (line.charAt(0) == '\t' || line.startsWith("$ELSEIF") || line.startsWith("$ELSE") || line.startsWith("$ENDIF"))
		{
		    currentLine.append(line);
		}
		else
		{
		    if (currentLine.length() > 0 && currentEntryName != null) // flush
		    {
			addDataToEntry(currentLine.toString(),currentEntryName);
			currentLine.setLength(0);
		    }
		    currentLine.setLength(0);
		    currentLine.append(line);
		}
	    }
	    
	    br.close();
	}
	catch (Exception e)
	{
	    System.out.println("The EseShiori Clone : exception occured in loading dictionary "+dicfile.getName());
	    e.printStackTrace();
	}
    }
    
    private void addDataToEntry(String data,String entry)
    {
	// ^u͏܂B
	// ^\n+,Ȃm擾܂B
	data = escape(data);
	
	if (data.indexOf('\t') != -1)
	{
	    StringBuffer buf = new StringBuffer(data);
	    while (true)
	    {
		int tab_pos = buf.toString().indexOf('\t');
		if (tab_pos == -1) break;
		
		buf.deleteCharAt(tab_pos);
	    }
	    data = buf.toString();
	}
	
	int prob = 1;
	if (data.indexOf(',') != -1)
	{
	    String probstr = data.substring(0,data.indexOf(','));
	    try
	    {
		prob = Integer.parseInt(probstr);
		data = data.substring(data.indexOf(',')+1);
	    }
	    catch (NumberFormatException e) {}
	}
	
	if (entry.charAt(0) == '#')
	{
	    // ^Cv𓾂
	    String type = entry.substring(1,5);
	    Vector target_entries = (Vector)special.get(type);
	    if (target_entries == null) // ߂Čꂽ^Cv
	    {
		target_entries = new Vector();
		special.put(type,target_entries);
	    }
	    
	    SCESSpecialEntry destination = null;
	    for (Enumeration e = target_entries.elements();e.hasMoreElements();)
	    {
		SCESSpecialEntry elem = (SCESSpecialEntry)e.nextElement();
		if (elem.getDefinition().equals(entry))
		{
		    destination = elem;
		}
	    }
	    if (destination == null) // ߂ČꂽGg
	    {
		destination = new SCESSpecialEntry(eseshiori,entry);
		target_entries.add(destination);
	    }
	    destination.add(prob,data);
	}
	else
	{
	    SCESNormalEntry destination = (SCESNormalEntry)root.get(entry);
	    if (destination == null)
	    {
		destination = new SCESNormalEntry();
		root.put(entry,destination);
	    }
	    destination.add(prob,data);
	}
    }
    
    public String getAnyItem(String entryname)
    {
	// GgChJ[hgp\B
	// mɏ]ĂǂꂩԂ܂B
	// GgȂ΋󔒕Ԃ܂B
	Vector entries = findEntries(entryname);
	if (entries.size() == 0) return "";
	
	// Ƃ肠GgIԁB
	int select_entry = (int)((Math.random() * 10000) % entries.size());
	SCESNormalEntry entry = (SCESNormalEntry)entries.elementAt(select_entry);
	
	String result = entry.getNextItem();
	return (result == null ? "" : result);
    }
    
    public String getItemMatches(String str,String type)
    {
	// str : }b`Op
	// type : RESP , EVNT , etc...
	Vector target_entries = (Vector)special.get(type);
	if (target_entries == null) return null; // ^Cvv̂Ȃ
	
	// [v̂𒲂ׂĂB
	// 炻ŏIB
	for (Enumeration e = target_entries.elements();e.hasMoreElements();)
	{
	    SCESSpecialEntry entry = (SCESSpecialEntry)e.nextElement();
	    if (entry.doMatchingWith(str))
	    {
		return entry.getNextItem(); // mlđI
	    }
	}
	return null; // vȂ
    }
    
    public Vector findEntries(String entryname)
    {
	// GgChJ[hgp\B
	// qbgGgSĕԂ܂B
	// ܂߂l̒gSCESNormalEntryB
	Vector result = new Vector();
	if (entryname.endsWith("*"))
	{
	    entryname = entryname.substring(0,entryname.indexOf('*'));
	}
	for (Enumeration e = root.keys();e.hasMoreElements();)
	{
	    String key = (String)e.nextElement();
	    if (key.startsWith(entryname))
	    {
		result.add(root.get(key));
	    }
	}
	return result;
    }
    
    public SCESNormalEntry findNormalEntry(String entry_name)
    {
	return (SCESNormalEntry)root.get(entry_name);
    }
    
    private static Hashtable _escape_table;
    static
    {
	_escape_table = new Hashtable();
	_escape_table.put("%%","&_percent_");
	_escape_table.put("%(","&_parenthesisopen_");
	_escape_table.put("%)","&_parenthesisclose_");
    }
    public static String escape(String src)
    {
	// ꕶGXP[v܂B
	// ł%&xxWJ܂B
	StringBuffer buf = new StringBuffer(src);
	while (true)
	{
	    int pos = buf.toString().indexOf("%&");//%&0a
	    if (pos == -1) break;
	    
	    try
	    {
		String hexstr = buf.substring(pos+2,pos+4);
		int code = Integer.parseInt(hexstr,16);
		buf.replace(pos,pos+4,""+(char)code);
	    }
	    catch (Exception e)
	    {
		buf.delete(pos,pos+2);
	    }
	}
	for (Enumeration e = _escape_table.keys();e.hasMoreElements();)
	{
	    String key = (String)e.nextElement();
	    int pos = buf.toString().indexOf(key);
	    if (pos == -1) continue;
	    
	    buf.replace(pos,pos+key.length(),(String)_escape_table.get(key));
	}
	return buf.toString();
    }
    
    private static Hashtable _unescape_table;
    static
    {
	_unescape_table = new Hashtable();
	_unescape_table.put("&_percent_","%");
	_unescape_table.put("&_parenthesisopen_","(");
	_unescape_table.put("&_parenthesisclose_",")");
	_unescape_table.put("&_dollar_","$"); // Fuckin'ʏ̐K@BSHIT!!
    }
    public static String unescape(String src)
    {
	StringBuffer buf = new StringBuffer(src);
	for (Enumeration e = _unescape_table.keys();e.hasMoreElements();)
	{
	    String key = (String)e.nextElement();
	    int pos = buf.toString().indexOf(key);
	    if (pos == -1) continue;
	    
	    buf.replace(pos,pos+key.length(),(String)_unescape_table.get(key));
	}
	return buf.toString();
    }
}
