/*
    "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
    Copyright (C) 2001 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import com.apple.cocoa.foundation.*;
import com.apple.cocoa.application.*;
import java.net.*;
import java.io.*;
import java.util.*;
import util.*;

public class SCServerServiceThread extends Thread {
    static final String sstp_version_str = "SSTP/1.4";
	
    static volatile boolean someone_speaking_on_sstp = false;
    
    Socket s;
    BufferedReader br;
    OutputStreamWriter osw;
    
    Vector tempEntries; // gStringB
    SCBalloonAlternative alternativeSelected;
	
    final boolean lightmode_on_sstp;
    final boolean reset_surfaces;
    
    public SCServerServiceThread(Socket sock) {
        s = sock;
        try {
            InputStream is = s.getInputStream();
            OutputStream os = s.getOutputStream();
	    br = new BufferedReader(new InputStreamReader(is));
            osw = new OutputStreamWriter(os,"SJIS");
	    
	    //System.out.println(s.getLocalAddress().toString()+" , "+s.getLocalAddress().equals(InetAddress.getLocalHost()));
	    /*boolean local = (s.getInetAddress().getHostAddress().equals("127.0.0.1") ||
	      s.getLocalAddress().equals(InetAddress.getLocalHost()));
	      System.out.println(local ? "local" : "external");*/
        } catch (Exception e) {}
	
	NSUserDefaults defaults = NSUserDefaults.standardUserDefaults();
	lightmode_on_sstp = (defaults.integerForKey("misc.lightmode_on_sstp") != 0);
	reset_surfaces = (defaults.integerForKey("misc.reset_surface_on_sstp") != 0);
    }
    
    public void run() {
	int pool = NSAutoreleasePool.push();
	
	try {
	    String header = null;
	    
	    SCPrefWindowController prefwindow = SCPrefWindowController.sharedPrefWindowController();
	    
	    header = br.readLine();
	    prefwindow.addStringToSSTPLog("> "+header+'\n');
	    
	    Hashtable request = new Hashtable();
	    String line = null;
	    while(true) {
		// sǂłANGXge[u쐬B
		try{  line = br.readLine();  }catch(Exception e){}
		if (line == null || line.equals("")) break;
		
		// SJISȊOŃNGXgꂽ烍O邪c
		prefwindow.addStringToSSTPLog(new String(("> "+line+'\n').getBytes(),"SJIS"));
		
		// Rŕ
		int colon_pos = line.indexOf(':');
		if (colon_pos == -1) {s.close(); NSAutoreleasePool.pop(pool); return;} // 肦ȂG[AIB
		
		String type = line.substring(0,colon_pos);
		String data = line.substring(colon_pos + 2); // R̎̃Xy[XB
		
		// SEND SSTP/1.4p̓ʏB
		// IfGhostȏ㗈RBƎdlOĂ܂B
		if (type.equals("IfGhost") && request.get(type) != null) {
		    writeResponceAndTerminate("501 Not Implemented");
		    NSAutoreleasePool.pop(pool);
		    return;
		}
		
		// SEND SSTP/1.2p̓ʏB
		// ZbVN\ꍇɃZbV̈ꎞGgɐςނ̂ŁA
		// ۑĂB
		if (type.equals("Entry")) {
		    if (tempEntries == null) tempEntries = new Vector();
		    tempEntries.add(data);
		}
		
		// nbVe[uɃZbgB
		request.put(type,data);
	    }
			
	    if (header.startsWith("SEND SSTP/")) {
		if (someone_speaking_on_sstp &&
		    SCFoundation.SSTP_CONFLICT_IF_ANYONE_SPEAKING){
			
		    writeResponceAndTerminate("409 Conflict");
		}
		else {
		    someone_speaking_on_sstp = true;
					
		    // ŗONsomeone_speaking_on_sstptruê܂܂ɂȂƁcB
		    try {
			doSSTPSend(request);
		    }
		    catch (Exception e) {
			e.printStackTrace();
		    }
					
		    someone_speaking_on_sstp = false;
		}
	    }
	    else if (header.startsWith("EXECUTE SSTP/")) {
		doSSTPExecute(request);
	    }
	    else if (header.startsWith("NOTIFY SSTP/")) {
		doSSTPNotify(request);
	    }
	    else {
		// Ή̃NGXgeB
		writeResponceAndTerminate("501 Not Implemented");
		NSAutoreleasePool.pop(pool);
		return;
	    }
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
		
	try{ s.close();  }catch(Exception ex){}
	NSAutoreleasePool.pop(pool);
    }
    
    private String getEncType(Hashtable request) {
	String enc_type = (String)request.get("Charset");
	if (enc_type == null) enc_type = "SJIS";
	
	// oJw΍
	if (enc_type.indexOf("EUC") != -1) enc_type = "EUC-JP";
	
	return enc_type;
    }
    
    private boolean hasAnyAlternatives(String scr) {
	// XNvg̒ɑIiq^OjłLtrueԂ܂B
	StringBuffer script = new StringBuffer(scr);
	
	// obNXbVɒu
	if (script.toString().indexOf('\\') != -1) {
	    int pos;
	    while ((pos = script.toString().indexOf('\\')) != -1) {
		script.setCharAt(pos,'\u00a5');
	    }
	}
	
	int search_offset = 0;
	while (search_offset < script.length()) {
	    int pos_enq = script.toString().indexOf("\u00a5q",search_offset);
	    if (pos_enq != -1) {
		if (pos_enq >= 1 && script.charAt(pos_enq-1) == '\u00a5') { // GXP[vĂ
		    // Á̒OGXP[v̓LZĂB
		    if (pos_enq >= 2 && script.charAt(pos_enq-2) == '\u00a5') {
			return true;
		    }
		    else {
			search_offset++;
		    }
		}
		else {
		    return true;
		}
	    }
	    else {
		return false;
	    }
	}
	return false;
    }
    
    private void writeResponceAndTerminate(String resp) {
	writeStr(sstp_version_str+' '+resp); writeCRLF();
	try{  s.close();  }catch(Exception e){}
	
	sstplog_out(resp);
    }
    
    private void writeStr(String param) {
	try{  osw.write(param,0,param.length());  }catch(Exception e){}
    }
    
    private void sstplog_out(String msg) {
	SCPrefWindowController prefwindow = SCPrefWindowController.sharedPrefWindowController();
	prefwindow.addStringToSSTPLog("< "+msg+"\n");
    }
    
    private void writeCRLF() {
	try {
	    osw.write(0x0d);
	    osw.write(0x0a);
	    osw.flush();
	} catch(Exception e) {}
    }
	
    protected void doSSTPSend(Hashtable request) throws UnsupportedEncodingException,IOException {
	if (request.get("Sender") == null || request.get("Script") == null) {
	    writeResponceAndTerminate("400 Bad Request");
	    return;
	}
	
	if (NSApplication.sharedApplication().isHidden()) {
	    writeResponceAndTerminate("512 Invisible");
	    return;
	}
	
	String enc_type = getEncType(request);
	String sender = new String( ((String)request.get("Sender")).getBytes() , enc_type );
	String script = new String( ((String)request.get("Script")).getBytes() , enc_type );
	
	Hashtable scriptrunner_options = new Hashtable();
	if (reset_surfaces) {
	    // \s[0]s[10]ɃZbg
	    scriptrunner_options.put("force-reset-surfaces-before-talk", new Boolean(true));
	    scriptrunner_options.put("show-partner-before-talk-if-not-used-yen-s", new Boolean(true));
	}
	scriptrunner_options.put(
	    "do-nothing-but-close-when-normal-alternative-selected", new Boolean(true));
	
	String ifghost_entry = (String)request.get("IfGhost");
	String ifghost = (ifghost_entry == null ? null : new String( ifghost_entry.getBytes() , enc_type ));
	
	SCFoundation foundation = SCFoundation.sharedFoundation();
	SCSession session;
	if (ifghost == null) {
	    // NĂZbV_ɑIԁB
	    session = foundation.getSession();
	}
	else if (lightmode_on_sstp) {
	    // ̃[hł́AYZbVNĂȂĂ
	    // ̃S[XgCXg[Ă
	    // Cg[hŋNB
	    session = foundation.getSession(ifghost,false);
	    if (session == null) {
		session = foundation.getTemporarySession(ifghost);
	    }
	}
	else {
	    // YZbVNĂȂ
	    // NĂZbV_ɑIԁB
	    session = foundation.getSession(ifghost,true);
	}
	
	if (session == null) { // ZbVNĂȂB
	    writeResponceAndTerminate("503 Service Unavailable");
	    return;
	}
	
	// ̃ZbVXNvgi[Œ蒆
	if (session.isScriptRunnerProtected()) {
	    writeResponceAndTerminate("409 Conflict");
	    return;
	}
	
	if (tempEntries != null) { // SEND SSTP/1.2 ꎞGgB̎_ŃZbVN\ł邱Ƃۏ؂Ă̂ŁB
	    int n_ents = tempEntries.size();
	    for (int i = 0;i < n_ents;i++) {
		String data = new String( ((String)tempEntries.elementAt(i)).getBytes() , enc_type);
		
		// ŏ̃J}ŕB
		int comma_pos = data.indexOf(',');
		if (comma_pos == -1) {
		    writeResponceAndTerminate("400 Bad Request");
		    return;
		}
		String ent_name = data.substring(0,comma_pos);
		String ent_script = data.substring(comma_pos+1);
		
		session.getTempScriptServer().addEntry(new SCScriptEntry(ent_name,ent_script));
	    }
	}

	if (hasAnyAlternatives(script)) {
	    // IL
	    SCScriptRunner runner = session.runScript(script,scriptrunner_options,false);
	    session.getHontaiBalloon().showSSTPMessage("from "+sender+" ("+s.getInetAddress().toString()+")");
	    alternativeSelected = null;
	    while (runner.getStatus() == SCScriptRunner.RUNNING) {
		try {
		    sleep((long)(0.25 * 1000)); // l̈buɃXe[^XB
		}
		catch (InterruptedException e) { break; }
	    }
	    alternativeSelected = runner.getSelectedAlternative();
	    
	    if (alternativeSelected == null) { // ^CAEgA^[~l[gꂽB
		// ZbVOnChoiceTimeouttHA[hB
		session.doShioriEvent("OnChoiceTimeout",new String[] {script});
		
		// 204ԂB
		writeStr(sstp_version_str+" 204 No Content"); writeCRLF();
		s.close();
		sstplog_out(sstp_version_str+" 204 No Content\n");
	    }
	    else {
		// http://Ŏn܂ĂURLJB
		if (SCScriptRunner.isURL(alternativeSelected.getRefcon())) {
		    //try{ NSWorkspace.sharedWorkspace().openURL(new URL(alternativeSelected.getRefcon())); }catch(Exception e){}
		    // URL͊runnerɊJĂB
		}
		else {
		    // łȂ΃GgXNvgƂċNB
		    session.runScript(alternativeSelected.getRefcon(),false);
		}
		
		// xNCAgɕԂB
		String label = alternativeSelected.title();
		writeStr(sstp_version_str+" 200 OK"); writeCRLF();
		writeCRLF();
		writeStr(label); writeCRLF();
		s.close();
		
		sstplog_out(sstp_version_str+" 200 OK");
		sstplog_out("");
		sstplog_out(label+"\n");
	    }
	    
	    session.getTempScriptServer().sweepAway(); // SSTP̈ꎞGg|B
	}
	else {
	    // I
	    writeStr(sstp_version_str+" 204 No Content"); writeCRLF();
	    s.close();
	    sstplog_out(sstp_version_str+" 204 No Content\n");
	    
	    SCScriptRunner runner = session.runScript(script,scriptrunner_options,false);
	    session.getHontaiBalloon().showSSTPMessage("from "+sender+" ("+s.getInetAddress().toString()+")");
	    
	    // I܂ő҂B
	    while (runner.getStatus() == SCScriptRunner.RUNNING) {
		try {
		    sleep((long)(0.25 * 1000)); // l̈buɃXe[^XB
		}
		catch (InterruptedException e) {
		    break;
		}
	    }
	}
    }
    
    protected void doSSTPExecute(Hashtable request) throws UnsupportedEncodingException,IOException {
	SCSession session = SCFoundation.sharedFoundation().getSession();
	if (session == null) { // ZbVNĂȂ
	    writeResponceAndTerminate("503 Service Unavailable");
	    return;
	}
	
	// EXECUTEo[Wԍ͖ʓ|Ȃ̂Ń`FbNȂB
	if (request.get("Sender") == null || request.get("Command") == null) {
	    writeResponceAndTerminate("400 Bad Request");
	    return;
	}
	
	String enc_type = getEncType(request);
	String sender = new String( ((String)request.get("Sender")).getBytes() , enc_type );
	String command = new String( ((String)request.get("Command")).getBytes() , enc_type );
	
	if (command.equals("GetName")) { // EXECUTE SSTP/1.0
	    String name = session.getMasterSpirit().getDescManager().getStrValue("name");
	    if (name == null) name = "";
	    
	    writeStr(sstp_version_str+" 200 OK"); writeCRLF();
	    writeCRLF();
	    writeStr(name); writeCRLF();
	    s.close();
	    sstplog_out(sstp_version_str+" 200 OK");
	    sstplog_out("");
	    sstplog_out(name+"\n");
	}
	else if (command.equals("GetVersion")) { // EXECUTE SSTP/1.2
	    writeStr(sstp_version_str+" 200 OK"); writeCRLF();
	    writeCRLF();
	    writeStr(SCFoundation.getVersion()); writeCRLF();
	    s.close();
	    sstplog_out(sstp_version_str+" 200 OK");
	    sstplog_out("");
	    sstplog_out(SCFoundation.getVersion()+"\n");
	}
	else {
	    writeResponceAndTerminate("501 Not Implemented");
	    return;
	}
    }
    
    protected void doSSTPNotify(Hashtable request) throws UnsupportedEncodingException,IOException {
	SCSession session = SCFoundation.sharedFoundation().getSession();
	if (session == null) { // ZbVNĂȂ
	    writeResponceAndTerminate("503 Service Unavailable");
	    return;
	}
	
	if (request.get("Sender") == null || request.get("Event") == null) {
	    writeResponceAndTerminate("400 Bad Request");
	    return;
	}
	
	String enc_type = getEncType(request);
	String sender = new String( ((String)request.get("Sender")).getBytes() , enc_type );
	String event = new String( ((String)request.get("Event")).getBytes() , enc_type );
	
	boolean local = (s.getInetAddress().getHostAddress().equals("127.0.0.1") ||
			 s.getLocalAddress().equals(InetAddress.getLocalHost()));
	
	StringBuffer shiorirequest = new StringBuffer();
	if (session.getMasterSpirit().getShioriProtocolVersion() == 2) {
	    shiorirequest.append("GET Sentence SHIORI/2.0\r\n");
	    shiorirequest.append("SecurityLevel: "+(local ? "local" : "external")+"\r\n");
	    shiorirequest.append("Sender: "+sender+"\r\n");
	    shiorirequest.append("Event: "+event+"\r\n");
	}
	else {
	    shiorirequest.append("GET SHIORI/3.0\r\n");
	    shiorirequest.append("SecurityLevel: "+(local ? "local" : "external")+"\r\n");
	    shiorirequest.append("Sender: "+sender+"\r\n");
	    shiorirequest.append("ID: "+event+"\r\n");
	}
	
	for (int i = 0;i <= 7;i++) {
	    String ref = (String)request.get("Reference"+i);
	    if (ref != null) {
		shiorirequest.append("Reference"+i+": "+(new String(ref.getBytes(),enc_type))+"\r\n");
	    }
	}
	shiorirequest.append("\r\n");
	session.forwardToShioriKernel(shiorirequest.toString());
	
	writeStr(sstp_version_str+" 204 No Content"); writeCRLF();
	s.close();
	sstplog_out(sstp_version_str+" 204 No Content\n");
    }

    protected void finalize() {
	//Logger.log(this, Logger.DEBUG, "finalized");
    }
}
