/*  $Id: SCSpirit.java,v 1.12 2004/04/08 08:56:26 phonohawk Exp $
	
    "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
    Copyright (C) 2001-2003 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
import com.apple.cocoa.foundation.*;
import com.apple.cocoa.application.*;
import java.io.*;
import java.util.*;
import util.*;

public class SCSpirit
{
    public static final int STAT_OK = 0; // ANZX\
    public static final int STAT_PREPARING = 1; // 
    
    protected volatile int spirit_status = STAT_PREPARING;
    SCSession session;
    
    SCDescription descManager;
    Vector makoto_mods;
    SCShioriAccessor shiori;
    int shiori_main_version; // 2 or 3
    int shiori_sub_version; // 0
    
    String dirname;
    File spiritRootDir;
    
    String name;
    String spirit_identification;
    String selfname;
    String keroname;
    
    public SCSpirit(SCSession session,String dirname) {
	this.session = session;
	this.dirname = dirname;
	
	_construct(session.getGhostDir(),dirname,null,session.isLightMode());
    }
    
    public SCSpirit(SCSession session,String dirname,SCStatusWindowController status) {
	this.session = session;
	this.dirname = dirname;
	
	_construct(session.getGhostDir(),dirname,status,session.isLightMode());
    }
    
    public SCSpirit(File ghost_root,String dirname,boolean light_mode) {
	this.session = null;
	this.dirname = dirname;
	
	_construct(ghost_root,dirname,null,light_mode);
    }

    public String toString() {
	StringBuffer buf = new StringBuffer();
	buf.append("session: {").append(session).append('}');
	buf.append("; shiori: {").append(shiori).append('}');
	buf.append("; shiori protocol version: ").append(shiori_main_version).
	    append('.').append(shiori_sub_version);
	if (makoto_mods.size() > 0) {
	    buf.append("; makoto: [");
	    for (Iterator ite = makoto_mods.iterator(); ite.hasNext(); ) {
		buf.append('{').append(ite.next()).append('}');
		if (ite.hasNext()) {
		    buf.append(", ");
		}
	    }
	    buf.append(']');
	}
	return buf.toString();
    }

    protected void finalize() {
	Logger.log(this, Logger.DEBUG, "finalized");
    }
    
    public void reloadAll() {
	// SHIORIMAKOTO[h܂B
	// ݂Ɏgׂ炸B
	spirit_status = STAT_PREPARING;

	SCSession session_backup = session;
	close();
	session = session_backup;
	_construct(session.getGhostDir(),dirname,null,session.isLightMode());
    }
    
    protected void _construct(File ghost_root,String dirname,SCStatusWindowController status,boolean light_mode) {
	spiritRootDir = new File(ghost_root,"ghost/"+dirname);
	descManager = new SCDescription(new File(spiritRootDir,"descript.txt"));
	//descManager.setTag(ghost_root.toString());
	
	name = descManager.getStrValue("name");
	selfname = descManager.getStrValue("sakura.name");
	keroname = descManager.getStrValue("kero.name");
	if (selfname == null) selfname = "";
	if (keroname == null) keroname = "";
	spirit_identification = selfname+','+keroname;
	
	printstat(status,"scope #0: "+(selfname.length() == 0 ? "none" : "%[color,white]"+selfname+"%[color,default]")+"\n");
	printstat(status,"scope #1: "+(keroname.length() == 0 ? "none" : "%[color,white]"+keroname+"%[color,default]")+"\n");
	
	// SHIORIW[N
	if (light_mode) {
	    // Cg[hłSHIORI[hȂB
	    shiori = null;
	}
	else {
	    printstat(status,"The spirit of the ghost,\n");
	    printstat(status,"%[color,orange]loading SHIORI subsystem... ");
	    long time_shiori = System.currentTimeMillis();
	    shiori = SCShioriLoader.loadShioriModule(spiritRootDir);

	    // SHIORI/2.xSHIORI/3.xGET VersionŔʁB
	    spirit_status = STAT_OK;
	    SCShioriSessionResponce getversion_resp = doShioriSession(
		"GET Version SHIORI/2.0\r\nSender: "+
		SCFoundation.STRING_FOR_SENDER+
		"\r\nSecurityLevel: local\r\n\r\n");
	    spirit_status = STAT_PREPARING;
	    if (getversion_resp.getHeader().startsWith("SHIORI/3.")) {
		shiori_main_version = 3;
	    }
	    else {
		shiori_main_version = 2;
	    }
	    shiori_sub_version = 0;
	    
	    printstat(status,"%[color,white]done.("+(System.currentTimeMillis()-time_shiori)+"ms)%[color,default]\n");
	}
	
	// MAKOTOW[N
	if (SCMakotoLoader.getModuleNames(spiritRootDir).size() == 0 ||
	    light_mode) {
	    // Cg[hłMAKOTO[hȂB
	    printstat(status,"no MAKOTO subsystems will be loaded.\n");
	    makoto_mods = new Vector();
	}
	else {
	    printstat(status,"%[color,orange]loading MAKOTO subsystems... ");
	    long time_makoto = System.currentTimeMillis();
	    makoto_mods = SCMakotoLoader.loadMakotoModules(spiritRootDir);
	    printstat(status,"%[color,white]done.("+(System.currentTimeMillis()-time_makoto)+"ms)%[color,default]\n");
	}
	
	spirit_status = STAT_OK;
	
	printstat(status,"This SHIORI subsystem uses protocol SHIORI/"+shiori_main_version+".0.\n");
    }
    protected void printstat(SCStatusWindowController status,String msg) {
	if (status != null) status.texttype_print(msg);
    }
    
    public int getShioriProtocolVersion() {
	return shiori_main_version;
    }
    
    private void shioriSessionLog(String data) {
	if (!SCFoundation.LOG_SHIORI_SESSION) return;
	
	BufferedReader br = new BufferedReader(new StringReader(data));
	String line = null;
	while (true) {
	    try{ line = br.readLine(); }catch (Exception e){}
	    if (line == null) break;
	    
	    try{ System.out.println(new String(line.getBytes("SJIS"))); }catch(Exception e){} 
	}
	
	try{ br.close(); }catch(IOException e){}
    }
    public synchronized SCShioriSessionResponce doShioriSession(String request) {
	// ̃\bhSpirit̎͏܂ŌĂяoubN܂̂łӉB
	
	if (shiori == null) {
	    // shiori[hĂȂ΁ASHIORI/2.0 500 Internal Server ErrorԂB
	    return new SCShioriSessionResponce("SHIORI/2.0 500 Internal Server Error",new Hashtable());
	}
	
	if (spirit_status == STAT_PREPARING) {
	    while (spirit_status != STAT_OK) {
		try {
		    Thread.currentThread().sleep(300);
		}
		catch (InterruptedException e) {
		    return null;
		}
	    }
	}
	BufferedReader br = new BufferedReader(new StringReader(shiori.request(request)));
	
	shioriSessionLog(request);
	
	// X|Xwb_擾
	String header = null;
	try{ header = br.readLine(); }catch(Exception e){}
	shioriSessionLog(header);
	
	// X|XŜǂ
	Hashtable hash = new Hashtable();
	String line = null;
	while (true) {
	    try{ line = br.readLine(); }catch (Exception e){}
	    if (line == null) break;
	    shioriSessionLog(line);
	    
	    if (line.indexOf(':') == -1) continue;
	    
	    String label = line.substring(0,line.indexOf(':'));
	    String data = line.substring(line.indexOf(':')+2); // R̎̃Xy[X͖
	    
	    hash.put(label,data);
	}
	
	try{ br.close(); }catch(IOException e){}
	
	return new SCShioriSessionResponce(header,hash);
    }
    
    public String getStringFromShiori(String id) {
	// SHIORIGET String𔭍s܂B
	// 擾ɎsnullԂ܂B
	SCShioriSessionResponce resp;
	if (getShioriProtocolVersion() == 3)
	    resp = doShioriSession("GET SHIORI/3.0\r\nSender: "+SCFoundation.STRING_FOR_SENDER+"\r\nID: "+id+"\r\nSecurityLevel: local\r\n\r\n");
	else
	    resp = doShioriSession("GET String SHIORI/2.5\r\nSender: "+SCFoundation.STRING_FOR_SENDER+"\r\nID: "+id+"\r\nSecurityLevel: local\r\n\r\n");
	
	if (resp == null) return null;
	if (resp.getHeader().indexOf("200 OK") == -1) return null;
	
	if (getShioriProtocolVersion() == 3)
	    return (String)resp.getResponce().get("Value");
	else
	    return (String)resp.getResponce().get("String");
    }
    
    public String translateWithShiori(String src) {
	if (src == null || src.length() == 0) return src;
	
	SCShioriSessionResponce resp;
	if (shiori_main_version == 3)
	    resp = doShioriSession("GET SHIORI/3.0\r\nSender: "+SCFoundation.STRING_FOR_SENDER+"\r\nID: OnTranslate\r\nReference0: "+src+"\r\nSecurityLevel: local\r\n\r\n");
	else
	    resp = doShioriSession("TRANSLATE Sentence SHIORI/2.6\r\nSender: "+SCFoundation.STRING_FOR_SENDER+"\r\nSentence: "+src+"\r\nSecurityLevel: local\r\n\r\n");
	if (resp == null) return src;
	if (resp.getHeader().indexOf("200 OK") == -1) return src;
	
	if (shiori_main_version == 3)
	    return (String)resp.getResponce().get("Value");
	else
	    return (String)resp.getResponce().get("Sentence");
    }
    
    public String translateWithMakoto(String src) {
	String current_str = src;
	
	final int n_makoto = makoto_mods.size();
	for (int i = 0;i < n_makoto;i++) {
	    current_str = translateWithMakoto(current_str,(SCMakotoAccessor)makoto_mods.elementAt(i));
	}
	
	return current_str;
    }
    
    protected void makotoSessionLog(String data) {
	if (!SCFoundation.LOG_MAKOTO_SESSION) return;
	
	BufferedReader br = new BufferedReader(new StringReader(data));
	String line = null;
	while (true) {
	    try{ line = br.readLine(); }catch (Exception e){}
	    if (line == null) break;
	    
	    try{ System.out.println(new String(line.getBytes("SJIS"))); }catch(Exception e){} 
	}
	
	try{ br.close(); }catch(IOException e){}
    }
    protected String translateWithMakoto(String src,SCMakotoAccessor makoto) {
	String request = "EXECUTE MAKOTO/2.0\r\nSender: "+SCFoundation.STRING_FOR_SENDER+"\r\nString: "+src+"\r\n";
	BufferedReader br = new BufferedReader(new StringReader(makoto.request(request)));
	makotoSessionLog(request);
	
	// X|Xwb_擾
	String header = null;
	try{ header = br.readLine(); }catch(Exception e){}
	makotoSessionLog(header);
	
	// X|XŜǂ
	Hashtable hash = new Hashtable();
	String line = null;
	while (true) {
	    try{ line = br.readLine(); }catch (Exception e){}
	    if (line == null) break;
	    makotoSessionLog(line);
	    
	    if (line.indexOf(':') == -1) continue;
	    
	    String label = line.substring(0,line.indexOf(':'));
	    String data = line.substring(line.indexOf(':')+2); // R̎̃Xy[X͖
	    
	    hash.put(label,data);
	}
	
	try{ br.close(); }catch(IOException e){}
	
	// 200 OKłȂsrĉ܂ܕԂAłȂString:̓eԂB
	return (header.indexOf("200 OK") != -1 ? (String)hash.get("String") : src);
    }
    
    public void close() {
	session = null;
	descManager = null;
	
	final int n_makoto = makoto_mods.size();
	for (int i = 0;i < n_makoto;i++) {
	    ((SCMakotoAccessor)makoto_mods.elementAt(i)).terminate();
	}
	makoto_mods.clear();

	if (shiori != null) {
	    shiori.terminate(); shiori = null;
	}
    }
    
    public SCDescription getDescManager() {
        return descManager;
    }
    
    public SCShioriAccessor getShiori() {
        return shiori;
    }
    
    public Vector getMakotoModules() {
        return makoto_mods;
    }
    
    public String getName() {
	return name;
    }
    
    public String getIdentification() {
        return spirit_identification;
    }
    
    public String getSelfName() {
	return selfname;
    }
    
    public String getKeroName() {
	return keroname;
    }
	
    public File getRootDir() {
	return spiritRootDir;
    }
}
