/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.eval;
import aya.*;

// 値には実数と文字列があります。

public class Value implements Cloneable
{
	protected static final int TYPE_INTEGER = 0;
	protected static final int TYPE_REAL = 1;
	protected static final int TYPE_STRING = 2;
	
	int type;
	long integer;
	double real;
	String str;
	
	public Value(long integer) {
		setInteger(integer);
	}
	
	public Value(double real) {
		setReal(real);
	}
	
	public Value(String str) {
		setString(str);
	}
	
	public Object clone() {
		if (isInteger()) {
			return new Value(integer);
		}
		else if (isReal()) {
			return new Value(real);
		}
		else {
			return new Value(str);
		}
	}
	
	public Value setInteger() {
		return setInteger(getInteger());
	}
	
	public Value setReal() {
		return setReal(getReal());
	}
	
	public Value setString() {
		return setString(getString());
	}
	
	public Value setInteger(long integer) {
		this.type = TYPE_INTEGER;
		this.integer = integer;
		this.real = 0;
		this.str = null;
		return this;
	}
	
	public Value setReal(double real) {
		if (real == Math.floor(real)) {
			setInteger((long)real);
		}
		else {
			this.type = TYPE_REAL;
			this.integer = 0;
			this.real = real;
			this.str = null;
		}
		return this;
	}
	
	public Value setString(String str) {
		this.type = TYPE_STRING;
		this.integer = 0;
		this.real = 0;
		this.str = str;
		return this;
	}
	
	public boolean isNumeric() {
		return type == TYPE_INTEGER || type == TYPE_REAL;
	}
	
	public boolean isInteger() {
		return type == TYPE_INTEGER;
	}
	
	public boolean isReal() {
		return type == TYPE_REAL;
	}
	
	public boolean isString() {
		return type == TYPE_STRING;
	}
	
	public long getInteger() {
		if (type == TYPE_INTEGER) {
			return integer;
		}
		else if (type == TYPE_REAL) {
			return (long)real;
		}
		else {
			// 文字列を整数に変換してみる。駄目なら0。
			try {
				return Long.parseLong(str);
			}
			catch (Exception e) {
				return 0;
			}
		}
	}
	
	public double getReal() {
		if (type == TYPE_INTEGER) {
			return (double)integer;
		}
		else if (type == TYPE_REAL) {
			return real;
		}
		else {
			// 文字列を実数に変換してみる。駄目なら0。
			try {
				return Double.parseDouble(str);
			}
			catch (Exception e) {
				return 0;
			}
		}
	}
	
	public String getString() {
		if (type == TYPE_STRING) {
			if (str == null) {
				return "";
			}
			else {
				return str;
			}
		}
		else if (type == TYPE_INTEGER) {
			return Long.toString(integer);
		}
		else {
			return Double.toString(real);
		}
	}
	
	public String toString() {
		if (type == TYPE_STRING) {
			return '"'+(str == null ? "" : str)+'"';
		}
		else if (type == TYPE_INTEGER) {
			return Long.toString(integer);
		}
		else {
			return Double.toString(real);
		}
	}
}
