/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.node;
import aya.*;
import aya.node.*;
import aya.eval.*;
import java.util.*;

public class Condition
{
	static protected Condition _factory = null;
	static public Condition getFactory() {
		if (_factory == null) {
			_factory = new Condition();
		}
		return _factory;
	}
	
	// 条件式は、一つ以上の部分条件式(Subcondition)と、
	// (部分条件式の個数)-1個の条件結合子から構成される。
	Vector subconditions; // [Subcondition,...]
	Vector combinations; // [String,...]
	
	protected Condition() {
		
	}
	
	public Condition(Subcondition cond) {
		subconditions = new Vector();
		combinations = null;
		
		subconditions.add(cond);
	}
	
	public Condition add(String combination,Subcondition cond) {
		if (combinations == null) {
			combinations = new Vector();
		}
		
		combinations.add(combination);
		subconditions.add(cond);
		return this;
	}
	
	public boolean eval(Namespace ns) throws Exception {
		boolean result = false;
		
		for (int i = 0; i < subconditions.size(); i++) {
			Subcondition subcond = (Subcondition)subconditions.elementAt(i);
			boolean bool = subcond.eval(ns);
			
			if (i == 0) {
				// 最初なので条件結合子は見ない。
				result = bool;
			}
			else {
				String cmb = (String)combinations.elementAt(i-1);
				if (cmb.equals("&&")) {
					result = result && bool;
				}
				else if (cmb.equals("||")) {
					result = result || bool;
				}
			}
		}
		
		return result;
	}
	
	public Subcondition newSubcondition(Condition cond) {
		return new ConditionalSubcondition(cond);
	}
	
	public Subcondition newSubcondition(Expression left,String comparison,Expression right) {
		return new ExpressionalSubcondition(left,comparison,right);
	}
	
	public class Subcondition {
		public boolean eval(Namespace ns) throws Exception {
			throw new Exception("Abstract method \"eval\" of Condition.Subcondition has been called directory.");
		}
	}
	
	public class ConditionalSubcondition extends Subcondition {
		// 条件式を持つ部分条件式。'(' 条件式 ')'で生成される。
		Condition cond;
		
		public ConditionalSubcondition(Condition cond) {
			this.cond = cond;
		}
		
		public boolean eval(Namespace ns) throws Exception {
			return cond.eval(ns);
		}
	}
	
	public class ExpressionalSubcondition extends Subcondition {
		// 二つの式(Expression)と一つの条件比較子から構成される部分条件式。
		Expression left,right;
		String oper;
		
		public ExpressionalSubcondition(Expression left,String oper,Expression right) {
			this.left = left;
			this.oper = oper;
			this.right = right;
		}
		
		public boolean eval(Namespace ns) throws Exception {
			// 条件比較子 := '==' | '!=' | '<' | '<=' | '>' | '>=' | '_in_' | '!_in_'
			Value lh = left.eval(ns);
			Value rh = right.eval(ns);
			
			// 左右で型が違う場合はエラー………にするのはやっぱりやめ。
			// 比較子が==か!=なら、それぞれ適切な結果を返す。
			// それ以外の比較子なら常に偽。
			if ((lh.isNumeric() && rh.isString()) || (lh.isString() && rh.isNumeric())) {
				if (oper.equals("==")) {
					return false;
				}
				else if (oper.equals("!=")) {
					return true;
				}
				else {
					//throw new Exception("Comparison of numeric value and string: "+left+' '+oper+' '+right+" (evaluated to "+lh+' '+oper+' '+rh+")");
					return false;
				}
			}
			
			if (oper.equals("==")) {
				if (lh.isNumeric()) {
					return lh.getReal() == rh.getReal();
				}
				else if (lh.isString()) {
					return lh.getString().equals(rh.getString());
				}
			}
			else if (oper.equals("!=")) {
				if (lh.isNumeric()) {
					return lh.getReal() != rh.getReal();
				}
				else if (lh.isString()) {
					return !lh.getString().equals(rh.getString());
				}
			}
			else if (oper.equals("<")) {
				if (lh.isNumeric()) {
					return lh.getReal() < rh.getReal();
				}
				else if (lh.isString()) {
					return lh.getString().compareTo(rh.getString()) < 0;
				}
			}
			else if (oper.equals("<=")) {
				if (lh.isNumeric()) {
					return lh.getReal() <= rh.getReal();
				}
				else if (lh.isString()) {
					return lh.getString().compareTo(rh.getString()) <= 0;
				}
			}
			else if (oper.equals(">")) {
				if (lh.isNumeric()) {
					return lh.getReal() > rh.getReal();
				}
				else if (lh.isString()) {
					return lh.getString().compareTo(rh.getString()) > 0;
				}
			}
			else if (oper.equals(">=")) {
				if (lh.isNumeric()) {
					return lh.getReal() >= rh.getReal();
				}
				else if (lh.isString()) {
					return lh.getString().compareTo(rh.getString()) >= 0;
				}
			}
			else if (oper.equals("_in_")) {
				if (lh.isNumeric()) {
					throw new Exception("Comparison of _in_ operator to numeric value.");
				}
				else {
					return rh.getString().indexOf(lh.getString()) != -1;
				}
			}
			else if (oper.equals("!_in_")) {
				if (lh.isNumeric()) {
					throw new Exception("Comparison of !_in_ operator to numeric value.");
				}
				else {
					return rh.getString().indexOf(lh.getString()) == -1;
				}
			}
			return false; // ここには来ないはず。
		}
	}
}
