/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.node;
import aya.*;
import aya.node.*;
import java.util.*;

public class FunctionCall
{
	protected static FunctionCall _factory = null;
	
	public static FunctionCall getFactory() {
		if (_factory == null) {
			_factory = new FunctionCall("");
		}
		return _factory;
	}

	String name; // 関数名
	Vector args; // [Argument,...] これがnullの場合は引数群そのものが無かった場合。
	
	public FunctionCall(String name) {
		this.name = name;
		this.args = null;
	}
	
	public String getName() {
		return name;
	}
	
	public Vector getArgs() {
		return args;
	}
	
	public FunctionCall prepareArgs() {
		args = new Vector();
		return this;
	}
	
	public FunctionCall addArg(Argument arg) {
		if (args == null) {
			args = new Vector();
		}
		args.add(arg);
		return this;
	}
	
	public Argument newArg(Expression expr) {
		return new ExpressionArgument(expr);
	}
	
	public Argument newArg(VariableReference vref) {
		return new ReferenceArgument(vref);
	}
	
	public class Argument {
	}
	
	public class ExpressionArgument extends Argument {
		Expression expr;
		public ExpressionArgument(Expression expr) {
			this.expr = expr;
		}
		public Expression getExpression() {
			return expr;
		}
	}
	
	public class ReferenceArgument extends Argument {
		VariableReference vref;
		public ReferenceArgument(VariableReference vref) {
			this.vref = vref;
		}
		public VariableReference getReference() {
			return vref;
		}
	}
}
