/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.node;
import aya.*;
import aya.node.*;
import aya.eval.*;
import java.util.*;

public class Substitution
{
	Aya aya;
	VariableReference vref;
	String oper; // 代入演算子またはインクリメント演算子。
	Expression expr; // 代入演算子が用いられた場合の、代入する式。
	
	public Substitution(Aya aya,VariableReference vref)
	{
		this.aya = aya;
		this.vref = vref;
		this.oper = null;
		this.expr = null;
	}
	
	public String toString() {
		if (expr == null) {
			return vref.toString()+oper;
		}
		else {
			return vref.toString()+' '+oper+' '+expr.toString();
		}
	}
	
	public Substitution setIncrementalOper(String oper)
	{
		this.oper = oper;
		return this;
	}
	
	public Substitution setSubstitutionalOper(String oper,Expression expr)
	{
		this.oper = oper;
		this.expr = expr;
		return this;
	}
	
	public void eval(Namespace ns) throws Exception {
		Namespace space = (vref.isGlobal() ? aya.getGlobalNamespace() : ns);
		Variable v = space.define(vref.getName());
		
		boolean index_exists = false;
		int index = -1;
		if (vref.getIndex() != null) {
			index_exists = true;
			index = vref.evalIndex(ns);
		}
		
		Value old_val;
		if (index_exists) {
			old_val = v.getValue(index);
		}
		else {
			old_val = v.getValue();
		}
		
		// old_valが文字列だった場合、次の演算子は許されない。
		// ++,--,-=,*=,/=,%=,+:=,-:=,*:=,/:=,%:=
		// 逆に言うと、許されるのは=,:=,+=のみ。
		// ただし、old_valが空文字列だった場合はゼロとして扱う。
		if (old_val.isString() &&
			!(oper.equals("=") || oper.equals(":=") || oper.equals("+="))) {
			
			if (old_val.getString().length() == 0) {
				old_val.setInteger(0);
			}
			else {
				throw new Exception("Illegal substitution: "+this.toString());
			}
		}
		
		if (oper.equals("++") || oper.equals("--")) {
			int signal = (oper.equals("++") ? 1 : -1);
			
			if (old_val.isInteger()) {
				old_val.setInteger(old_val.getInteger() + signal);
			}
			else if (old_val.isReal()) {
				old_val.setReal(old_val.getReal() + signal);
			}
		}
		else {
			Value val = expr.eval(ns);
			if (val == null) {
				val = new Value("");
			}
			
			if (val.isNumeric() && oper.indexOf(':') == -1) {
				// 値が数値であり、且つコロンを含まない演算子なので、小数点以下を切り捨て。
				val.setInteger(val.getInteger());
			}
			
			if (oper.equals("=") || oper.equals(":=")) {
				if (val.isInteger()) {
					old_val.setInteger(val.getInteger());
				}
				else if (val.isReal()) {
					old_val.setReal(val.getReal());
				}
				else if (val.isString()) {
					old_val.setString(val.getString());
				}
			}
			else if (oper.equals("+=") || oper.equals("+:=")) {
				if (val.isInteger()) {
					old_val.setInteger(old_val.getInteger() + val.getInteger());
				}
				else if (val.isReal()) {
					old_val.setReal(old_val.getReal() + val.getReal());
				}
				else if (val.isString()) {
					old_val.setString(old_val.getString() + val.getString());
				}
			}
			else if (oper.equals("-=") || oper.equals("-:=")) {
				if (val.isInteger()) {
					old_val.setInteger(old_val.getInteger() - val.getInteger());
				}
				else if (val.isReal()) {
					old_val.setReal(old_val.getReal() - val.getReal());
				}
			}
			else if (oper.equals("*=") || oper.equals("*:=")) {
				if (val.isInteger()) {
					old_val.setInteger(old_val.getInteger() * val.getInteger());
				}
				else if (val.isReal()) {
					old_val.setReal(old_val.getReal() * val.getReal());
				}
			}
			else if (oper.equals("/=") || oper.equals("/:=")) {
				if (val.isInteger()) {
					old_val.setInteger(old_val.getInteger() / val.getInteger());
				}
				else if (val.isReal()) {
					old_val.setReal(old_val.getReal() / val.getReal());
				}
			}
			else if (oper.equals("%=") || oper.equals("%:=")) {
				if (val.isInteger()) {
					old_val.setInteger(old_val.getInteger() % val.getInteger());
				}
				else if (val.isReal()) {
					old_val.setReal(old_val.getReal() % val.getReal());
				}
			}
		}
		
		if (index_exists) {
			v.setValue(index,old_val);
		}
		else {
			v.setValue(old_val);
		}
	}
}
