/*
    "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
    Copyright (C) 2001 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package satori;
import satori.*;
import java.util.*;

public class LinedDataBlock {
    public static final int TYPE_SENTENCE = 1; // ̏W
    public static final int TYPE_WORD = 2; // P̏W

    public static final int METHOD_DISABLED = 1; // : 
    public static final int METHOD_NONREPEAT = 2; // : O
    public static final int METHOD_RANDOMIZED = 3; // : L
    public static final int METHOD_ASC = 4; // : 
    public static final int METHOD_DESC = 5; // : ~

    public static final int PERIOD_TALK = 1; // : g[N
    public static final int PERIOD_MATERIALIZED = 2; // : N
    
    Variable var; // ϐ̏WBQƂ邽߁c
    String name; // ubNB⁗͕tȂB
    int type; // ̃ubÑ^CvB
    Vector data; // [String,...] f[^

    int nonoverlap_method;
    int nonoverlap_period;
    
    Vector remaining; // [String,...] _ȏdgĂꍇ̟kB
    int last_selected; // AȂdgĂꍇ́AOI񂾍ځB
    // ~̐ɂpB

    public LinedDataBlock(Variable var, String name, int type) {
	this.var = var;
	this.name = name;
	this.type = type;
	this.data = new Vector();
	this.nonoverlap_method = 0;
	this.nonoverlap_period = 0;
	this.remaining = null;
	this.last_selected = -1;
    }

    public int size() {
	return data.size();
    }

    public void add(String line) {
	data.add(line);
    }

    public Vector getContents() {
	return data;
    }

    public Enumeration elements() {
	return data.elements();
    }

    public String getFirst() {
	if (data.size() == 0) {
	    return null;
	}
	else {
	    return (String)data.get(0);
	}
    }

    public String get() {
	updateSetting();

	switch (nonoverlap_method) {
	case METHOD_DISABLED:
	    return getNormally();

	case METHOD_NONREPEAT:
	    return getNonrepeatly();

	case METHOD_RANDOMIZED:
	    return getRandomly();

	case METHOD_ASC:
	    return getAscently();

	case METHOD_DESC:
	    return getDescently();

	default:
	    return getNormally();
	}
    }

    public void updateSetting() {
	// (|PQ)u(<ubN>|)v̏d
	StringBuffer buf = new StringBuffer();
	buf.append(type == TYPE_SENTENCE ? ""+Satori.ZENKAKU_BUN : Satori.ZENKAKU_TANGOGUN);
	buf.append(Satori.ZENKAKU_KAGIKAKKO_OPEN);
	buf.append(name);
	buf.append(Satori.ZENKAKU_KAGIKAKKO_CLOSE);
	buf.append(Satori.ZENKAKU_NO_JUUHUKUKAIHI);
	String setting = var.get(buf.toString());
	
	if (setting == null) {
	    buf.setLength(0);
	    buf.append(type == TYPE_SENTENCE ? ""+Satori.ZENKAKU_BUN : Satori.ZENKAKU_TANGOGUN);
	    buf.append(Satori.ZENKAKU_KAGIKAKKO_OPEN);
	    buf.append('*');
	    buf.append(Satori.ZENKAKU_KAGIKAKKO_CLOSE);
	    buf.append(Satori.ZENKAKU_NO_JUUHUKUKAIHI);
	    setting = var.get(buf.toString());
	}

	interpretSetting(setting == null ? "" : setting);
    }

    protected static Hashtable setting_cache = null; // {ݒ蕶 => (int)[, ]}
    protected void interpretSetting(String setting) {
	if (setting_cache == null) {
	    setting_cache = new Hashtable();
	}
	else {
	    int[] cached = (int[])setting_cache.get(setting);
	    if (cached != null) {
		nonoverlap_method = cached[0];
		nonoverlap_period = cached[1];
		return;
	    }
	}

	int touten_pos = setting.indexOf(Satori.ZENKAKU_TOUTEN);
	String method, period;
	if (touten_pos == -1) {
	    method = setting;
	    period = "";
	}
	else {
	    method = setting.substring(0, touten_pos);
	    period = setting.substring(touten_pos+1);
	}

	if (method.equals(Satori.ZENKAKU_MUKOU)) {
	    nonoverlap_method = METHOD_DISABLED;
	}
	else if (method.equals(Satori.ZENKAKU_CHOKUZEN)) {
	    nonoverlap_method = METHOD_NONREPEAT;
	}
	else if (method.equals(Satori.ZENKAKU_YUUKOU)) {
	    nonoverlap_method = METHOD_RANDOMIZED;
	}
	else if (method.equals(Satori.ZENKAKU_SHOUJUN)) {
	    nonoverlap_method = METHOD_ASC;
	}
	else if (method.equals(Satori.ZENKAKU_KOUJUN)) {
	    nonoverlap_method = METHOD_DESC;
	}
	else {
	    nonoverlap_method =
		type == TYPE_SENTENCE ? METHOD_DISABLED : METHOD_RANDOMIZED;
	}

	if (period.equals(Satori.ZENKAKU_TALK_CHUU)) {
	    nonoverlap_period = PERIOD_TALK;
	}
	else if (period.equals(Satori.ZENKAKU_KIDOU_CHUU)) {
	    nonoverlap_period = PERIOD_MATERIALIZED;
	}
	else {
	    nonoverlap_period =
		type == TYPE_SENTENCE ? PERIOD_MATERIALIZED : PERIOD_TALK;
	}

	setting_cache.put(
	    setting,
	    new int[] {nonoverlap_method, nonoverlap_period});
    }

    protected String getNormally() {
	if (data.size() == 0) {
	    return null;
	}
	else {
	    int idx = (int)(Math.random() * Integer.MAX_VALUE) % data.size();
	    return (String)data.get(idx);
	}
    }

    protected String getRandomly() {
	if (remaining == null) {
	    remaining = new Vector();
	}
	
	if (remaining.size() == 0) {
	    // ď[Us
	    remaining.addAll(data);
	}

	if (remaining.size() == 0) {
	    return null;
	}
	else {
	    int idx = (int)(Math.random() * Integer.MAX_VALUE) % remaining.size();
	    return (String)remaining.remove(idx);
	}
    }

    protected String getNonrepeatly() {
	if (data.size() == 0) {
	    last_selected = -1;
	    return null;
	}
	else if (data.size() == 1) {
	    last_selected = 0;
	    return (String)data.get(0);
	}
	
	int id = -1;
	for (int count = 0; count < 100 && id != last_selected; count++) {
	    id = (int)(Math.random() * Integer.MAX_VALUE) % data.size();
	}

	last_selected = id;
	return (String)data.get(id);
    }

    protected String getAscently() {
	if (data.size() == 0) {
	    last_selected = -1;
	    return null;
	}

	last_selected++;
	if (last_selected >= data.size()) {
	    last_selected = 0;
	}

	return (String)data.get(last_selected);
    }

    protected String getDescently() {
	if (data.size() == 0) {
	    last_selected = -1;
	    return null;
	}

	last_selected--;
	if (last_selected < 0) {
	    last_selected = data.size()-1;
	}

	return (String)data.get(last_selected);
    }
}